<?php
/************************************************************************
 * MXChange v0.2.1                                    Start: 08/29/2004 *
 * ===============                              Last change: 08/29/2004 *
 *                                                                      *
 * -------------------------------------------------------------------- *
 * File              : lib-mysql3.php                                   *
 * -------------------------------------------------------------------- *
 * Short description : Database layer for MySQL +3.x server             *
 * -------------------------------------------------------------------- *
 * Kurzbeschreibung  : Datenbankschicht fuer MySQL +3.x Server          *
 * -------------------------------------------------------------------- *
 * $Revision::                                                        $ *
 * $Date::                                                            $ *
 * $Tag:: 0.2.1-FINAL                                                 $ *
 * $Author::                                                          $ *
 * Needs to be in all Files and every File needs "svn propset           *
 * svn:keywords Date Revision" (autoprobset!) at least!!!!!!            *
 * -------------------------------------------------------------------- *
 * Copyright (c) 2003 - 2009 by Roland Haeder                           *
 * For more information visit: http://www.mxchange.org                  *
 *                                                                      *
 * This program is free software; you can redistribute it and/or modify *
 * it under the terms of the GNU General Public License as published by *
 * the Free Software Foundation; either version 2 of the License, or    *
 * (at your option) any later version.                                  *
 *                                                                      *
 * This program is distributed in the hope that it will be useful,      *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of       *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *
 * GNU General Public License for more details.                         *
 *                                                                      *
 * You should have received a copy of the GNU General Public License    *
 * along with this program; if not, write to the Free Software          *
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,               *
 * MA  02110-1301  USA                                                  *
 ************************************************************************/

// Some security stuff...
if (!defined('__SECURITY')) {
	die();
} // END - if

// SQL queries
function SQL_QUERY ($sqlString, $F, $L) {
	// Trim SQL string
	$sqlString = trim($sqlString);

	// Link is up?
	if (!SQL_IS_LINK_UP()) {
		// We should not quietly ignore this!
		debug_report_bug(sprintf("Cannot query database: sqlString=%s,file=%s,line=%s",
			$sqlString,
			basename($F),
			$L
		));

		// Return 'false' because it has failed
		return false;
	} elseif (empty($sqlString)) {
		// Empty SQL string!
		debug_report_bug(sprintf("SQL string is empty. Please fix this. file=%s, line=%s",
			basename($F),
			$L
		));

		// This is invalid, of course
		return false;
	}

	// Remove \t, \n and \r from queries they may confuse some MySQL version I have heard
	$sqlString = str_replace("\t", ' ', str_replace("\n", ' ', str_replace("\r", ' ', $sqlString)));

	// Compile config out
	$sqlString = FILTER_COMPILE_CONFIG($sqlString);

	// Starting time
	$querytimeBefore = array_sum(explode(' ', microtime()));

	// Run SQL command
	///* DEBUG: */ print($sqlString . '<br />');
	$result = mysql_query($sqlString, SQL_GET_LINK())
		or addFatalMessage(__FUNCTION__, __LINE__, $F . ' (' . $L . '):' . mysql_error() . '<br />
Query string:<br />
' . $sqlString);
	//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'sql=' . $sqlString . ',numRows=' . SQL_NUMROWS($result) . ',affected=' . SQL_AFFECTEDROWS());

	// Ending time
	$querytimeAfter = array_sum(explode(' ', microtime()));

	// Calculate query time
	$queryTime = $querytimeAfter - $querytimeBefore;

	// Add this query to array including timing
	addSqlToDebug($result, $sqlString, $queryTime, $F, $L);

	// Save last successfull query
	setConfigEntry('db_last_query', $sqlString);

	// Count all query times
	incrementConfigEntry('sql_time', $queryTime);

	// Count this query
	incrementConfigEntry('sql_count');

	// Debug output
	if ((getOutputMode() != 1) && (isDebugModeEnabled()) && (isConfigEntrySet('DEBUG_SQL')) && (getConfig('DEBUG_SQL') == 'Y')) {
		//
		// Debugging stuff...
		//
		$fp = fopen(getConfig('CACHE_PATH') . 'mysql.log', 'a') or app_die(__FILE__, __LINE__, "Cannot write mysql.log!");
		if (!isset($GLOBALS['sql_first_entry'])) {
			// Write first entry
			fwrite($fp, 'Module=' . getModule() . "\n");
			$GLOBALS['sql_first_entry'] = true;
		} // END - if
		fwrite($fp, $F . '(LINE=' . $L . '|NUM=' . SQL_NUMROWS($result) . '|AFFECTED=' . SQL_AFFECTEDROWS() . '|QUERYTIME:' . $queryTime . '): ' . str_replace("\r", '', str_replace("\n", ' ', $sqlString)) . "\n");
		fclose($fp);
	} // END - if

	// Count DB hits
	if (!isStatsEntrySet('db_hits')) {
		// Count in dummy variable
		setStatsEntry('db_hits', 1);
	} else {
		// Count to config array
		incrementStatsEntry('db_hits');
	}

	// Return the result
	return $result;
}

// SQL num rows
function SQL_NUMROWS ($result) {
	// Link is not up, no rows by default
	$lines = false;

	// Is the result a valid resource?
	if (is_resource($result)) {
		// Get the count of rows from database
		$lines = mysql_num_rows($result);

		// Is the result empty? Then we have an error!
		if (empty($lines)) $lines = 0;
	} elseif (SQL_IS_LINK_UP()) {
		// No resource given, no lines found!
		$lines = 0;
	}

	// Return lines
	return $lines;
}

// SQL affected rows
function SQL_AFFECTEDROWS() {
	// Valid link resource?
	if (!SQL_IS_LINK_UP()) return false;

	// Get affected rows
	$lines = mysql_affected_rows(SQL_GET_LINK());

	// Return it
	return $lines;
}

// SQL fetch row
function SQL_FETCHROW ($result) {
	// Is a result resource set?
	if ((!is_resource($result)) || (!SQL_IS_LINK_UP())) return false;

	// Fetch the data and return it
	return mysql_fetch_row($result);
}

// SQL fetch array
function SQL_FETCHARRAY ($res, $nr=0, $remove_numerical=true) {
	// Is a result resource set?
	if ((!is_resource($res)) || (!SQL_IS_LINK_UP())) return false;

	// Initialize array
	$row = array();

	// Load row from database
	$row = mysql_fetch_array($res);

	// Return only arrays here
	if (is_array($row)) {
		// Shall we remove numerical data here automatically?
		if ($remove_numerical) {
			// So let's remove all numerical elements to save memory!
			$max = count($row);
			for ($idx = 0; $idx < ($max / 2); $idx++) {
				// Remove entry
				unset($row[$idx]);
			} // END - for
		} // END - if

		// Return row
		return $row;
	} else {
		// Return a false here...
		return false;
	}
}

// SQL result
function SQL_RESULT ($res, $row, $field = 0) {
	// Is $res valid?
	if ((!is_resource($res)) || (!SQL_IS_LINK_UP())) return false;

	// Run the result command and return the result
	$result = mysql_result($res, $row, $field);
	return $result;
}

// SQL connect
function SQL_CONNECT ($host, $login, $password, $F, $L) {
	// Try to connect
	$connect = mysql_connect($host, $login, $password) or addFatalMessage(__FUNCTION__, __LINE__, $F." (".$L."):".mysql_error());

	// Set the link resource
	SQL_SET_LINK($connect);

	// Destroy cache
	unset($GLOBALS['sql_link_res']);
}

// SQL select database
function SQL_SELECT_DB ($dbName, $F, $L) {
	// Is there still a valid link? If not, skip it.
	if (!SQL_IS_LINK_UP()) return false;

	// Return the result
	return mysql_select_db($dbName, SQL_GET_LINK()) or addFatalMessage(__FUNCTION__, __LINE__, $F." (".$L."):".mysql_error());
}

// SQL close link
function SQL_CLOSE ($F, $L) {
	if (!SQL_IS_LINK_UP()) {
		// Skip double close
		return false;
	} // END - if

	// Close database link and forget the link
	$close = mysql_close(SQL_GET_LINK())
		or addFatalMessage(__FUNCTION__, __LINE__, $F . ' (' . $L . '):'.mysql_error());

	// Close link
	SQL_SET_LINK(null);

	// Destroy cache
	unset($GLOBALS['sql_link_res']);

	// Return the result
	return $close;
}

// SQL free result
function SQL_FREERESULT ($result) {
	if ((!is_resource($result)) || (!SQL_IS_LINK_UP())) {
		// Abort here
		return false;
	} // END - if

	$res = mysql_free_result($result);
	return $res;
}

// SQL string escaping
function SQL_QUERY_ESC ($qstring, $data, $F, $L, $run=true, $strip=true, $secure=true) {
	// Link is there?
	if ((!SQL_IS_LINK_UP()) || (!is_array($data))) return false;

	// Escape all data
	$dataSecured['__sql_string'] = $qstring;
	foreach ($data as $key => $value) {
		$dataSecured[$key] = SQL_ESCAPE($value, $secure, $strip);
	} // END - foreach

	// Generate query
	$query = call_user_func_array('sprintf', $dataSecured);

	// Debugging
	//
	///* DEBUG: */ $fp = fopen(getConfig('CACHE_PATH') . 'escape_debug.log', 'a') or app_die(__FILE__, __LINE__, "Cannot write debug.log!");
	///* DEBUG: */ fwrite($fp, $F.'('.$L."): ".str_replace("\r", '', str_replace("\n", " ", $eval))."\n");
	///* DEBUG: */ fclose($fp);

	if ($run === true) {
		// Run SQL query (default)
		return SQL_QUERY($query, $F, $L);
	} else {
		// Return secured string
		return $query;
	}
}

// Get id from last INSERT command
function SQL_INSERTID () {
	if (!SQL_IS_LINK_UP()) return false;
	return mysql_insert_id();
}

// Escape a string for the database
function SQL_ESCAPE ($str, $secureString=true, $strip=true) {
	// Do we have cache?
	if (!isset($GLOBALS['sql_escapes'][''.$str.''])) {
		// Secure string first? (which is the default behaviour!)
		if ($secureString === true) {
			// Then do it here
			$str = secureString($str, $strip);
		} // END - if

		if (!SQL_IS_LINK_UP()) {
			// Fall-back to smartAddSlashes() when there is no link
			$ret = smartAddSlashes($str);
		} elseif (function_exists('mysql_real_escape_string')) {
			// The new and improved version
			///* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'str='.$str);
			$ret = mysql_real_escape_string($str, SQL_GET_LINK());
		} elseif (function_exists('mysql_escape_string')) {
			// The obsolete function
			$ret = mysql_escape_string($str, SQL_GET_LINK());
		} else {
			// If nothing else works, fall back to smartAddSlashes() again
			$ret = smartAddSlashes($str);
		}

		// Cache result
		$GLOBALS['sql_escapes'][''.$str.''] = $ret;
	} // END - if

	// Return it
	return $GLOBALS['sql_escapes'][''.$str.''];
}

// SELECT query string from table, columns and so on... ;-)
function SQL_RESULT_FROM_ARRAY ($table, $columns, $idRow, $id, $F, $L) {
	// Is columns an array?
	if (!is_array($columns)) {
		// No array
		debug_report_bug(sprintf("columns is not an array. %s != array, file=%s, line=%s",
			gettype($columns),
			basename($F),
			$L
		));

		// Abort here with 'false'
		return false;
	} // END  - if

	// Prepare the SQL statement
	$sql = "SELECT `".implode("`,`", $columns)."` FROM `{?_MYSQL_PREFIX?}_%s` WHERE `%s`='%s' LIMIT 1";

	// Return the result
	return SQL_QUERY_ESC($sql,
		array(
			$table,
			$idRow,
			bigintval($id),
		), $F, $L
	);
}

// ALTER TABLE wrapper function
function SQL_ALTER_TABLE ($sql, $F, $L) {
	// Abort if link is down
	if (!SQL_IS_LINK_UP()) return false;

	// This is the default result...
	$result = false;

	// Determine index/fulltext/unique word
	$noIndex = (
	(
		strpos($sql, 'INDEX') === false
	) && (
		strpos($sql, 'KEY') === false
	) && (
		strpos($sql, 'FULLTEXT') === false
	) && (
		strpos($sql, 'UNIQUE') === false
	)
	);

	// Extract table name
	$tableArray = explode(' ', $sql);
	$tableName = str_replace('`', '', $tableArray[2]);

	// Shall we add/drop?
	if (((strpos($sql, 'ADD') !== false) || (strpos($sql, 'DROP') !== false)) && ($noIndex === true)) {
		// Try two columns, one should fix
		foreach (array(4,5) as $idx) {
			// And column name as well
			$columnName = str_replace('`', '', $tableArray[$idx]);

			// Get column information
			$result = SQL_QUERY_ESC("SHOW COLUMNS FROM `%s` LIKE '%s'",
				array($tableName, $columnName), __FILE__, __LINE__);

			// Do we have no entry on ADD or an entry on DROP?
			// 123           4       4     3    3      4           4          32    23           4       4     3    3      4            4          321
			if (((SQL_NUMROWS($result) == 0) && (strpos($sql, 'ADD') !== false)) || ((SQL_NUMROWS($result) == 1) && (strpos($sql, 'DROP') !== false))) {
				// Do the query
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Executing: ' . $sql);
				$result = SQL_QUERY($sql, $F, $L, false);

				// Skip further attempt(s)
				break;
			} elseif ((((SQL_NUMROWS($result) == 1) && (strpos($sql, 'ADD') !== false)) || ((SQL_NUMROWS($result) == 0) && (strpos($sql, 'DROP') !== false))) && ($columnName != 'KEY')) {
				// Abort here because it is alreay there
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Skipped: ' . $sql);
				break;
			} elseif ($columnName != 'KEY') {
				// Something didn't fit
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Possible problem: ' . $sql);
			}
		} // END - foreach
	} elseif ((getConfig('_TABLE_TYPE') == 'InnoDB') && (strpos($sql, 'FULLTEXT') !== false)) {
		// Skip this query silently
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("Skipped FULLTEXT: sql=%s,file=%s,line=%s", $sql, $F, $L));
	} elseif ($noIndex === false) {
		// And column name as well
		//* DEBUG: */ print __LINE__.':tableArray=<pre>' . print_r($tableArray, true) . '</pre>';
		$keyName = str_replace('`', '', $tableArray[5]);

		// Is this "UNIQUE" or so? FULLTEXT has been handled the elseif() block above
		if (in_array(strtoupper($keyName), array('INDEX', 'UNIQUE', 'KEY', 'FULLTEXT'))) {
			// Init loop
			$begin = 1; $keyName = ',';
			while (strpos($keyName, ',') !== false) {
				// Use last
				$keyName = str_replace('`', '', $tableArray[count($tableArray) - $begin]);
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $keyName . '----------------' . $begin);

				// Remove brackes
				$keyName = str_replace('(', '', str_replace(')', '', $keyName));
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $keyName . '----------------' . $begin);

				// Continue
				$begin++;
			} // END while
		} // END - if

		// Show indexes
		$result = SQL_QUERY_ESC("SHOW INDEX FROM `%s`", array($tableName), __FILE__, __LINE__);

		// Non-skipping is default for ADD
		$skip = false;

		// But should we DROP?
		if ($tableArray[3] == 'DROP') {
			// Then skip if nothing found!
			$skip = true;
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Going to drop key ' . $keyName);
		} // END - if

		// Walk through all
		while ($content = SQL_FETCHARRAY($result)) {
			// Is it found?
			//* DEBUG: */ print(__LINE__.':columnName='.$keyName.',content=<pre>' . print_r($content, true) . '</pre>');
			if (($content['Key_name'] == $keyName) && ($tableArray[3] == 'ADD')) {
				// Skip this query!
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("ADD: Skiped: %s", $sql));
				$skip = true;
				break;
			} elseif (($content['Key_name'] == $keyName) && ($tableArray[3] == 'DROP')) {
				// Don't skip this!
				//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, sprintf("DROP: Not skiped: %s", $sql));
				$skip = false;
				break;
			}
		} // END - while

		// Free result
		SQL_FREERESULT($result);

		// Shall we run it?
		if ($skip === false) {
			// Send it to the SQL_QUERY() function
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $sql);
			$result = SQL_QUERY($sql, $F, $L, false);
		} else {
			// Not executed
			//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, 'Not executed: ' . $sql);
		}
	} else {
		// Other ALTER TABLE query
		//* DEBUG: */ logDebugMessage(__FUNCTION__, __LINE__, $sql);
		$result = SQL_QUERY($sql, $F, $L, false);
	}

	// Return result
	return $result;
}

// Getter for SQL link
function SQL_GET_LINK () {
	// Init link
	$link = null;

	// Is it in the globals?
	if (isset($GLOBALS['sql_link'])) {
		// Then take it
		$link = $GLOBALS['sql_link'];
	} // END - if

	// Return it
	return $link;
}

// Setter for link
function SQL_SET_LINK ($link) {
	// Is this a resource or null?
	if ((!is_resource($link)) && (!is_null($link))) {
		// This should never happen!
		debug_report_bug(sprintf("link is not resource or null. Type: %s", gettype($link)));
	} // END - if

	// Set it
	$GLOBALS['sql_link'] = $link;
}

// Checks if the link is up
function SQL_IS_LINK_UP () {
	// Default is not up
	$linkUp = false;

	// Do we have cached this?
	if (isset($GLOBALS['sql_link_res'])) {
		// Then use this
		$linkUp = $GLOBALS['sql_link_res'];
	} else {
		// Get it
		$linkUp = is_resource(SQL_GET_LINK());

		// And cache it
		$GLOBALS['sql_link_res'] = $linkUp;
	}

	// Return the result
	return $linkUp;
}

// [EOF]
?>
